"""Django app for accessing the EVE Swagger Interface (ESI)."""

__version__ = "8.2.5"
__title__ = 'Django-ESI'
__url__ = 'https://gitlab.com/allianceauth/django-esi'
__build_date__ = "2025-12-16"
__esi_compatibility_date__ = "2025-12-16"
