# Generated by Django 3.2.12 on 2022-02-19 18:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0010_set_new_tokens_to_sso_v2'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='character_id',
            field=models.IntegerField(db_index=True, help_text='The ID of the EVE character who authenticated by SSO.'),
        ),
        migrations.AlterField(
            model_name='token',
            name='character_name',
            field=models.CharField(db_index=True, help_text='The name of the EVE character who authenticated by SSO.', max_length=100),
        ),
        migrations.AlterField(
            model_name='token',
            name='character_owner_hash',
            field=models.CharField(db_index=True, help_text='The unique string identifying this character and its owning EVE account. Changes if the owning account changes.', max_length=254),
        ),
    ]
