# Generated by Django 3.2.12 on 2022-05-16 17:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0011_add_token_indices'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='token_type',
            field=models.CharField(choices=[('character', 'Character'), ('corporation', 'Corporation')], default='character', help_text='The applicable range of the token.', max_length=100),
        ),
    ]
